
#include <ctype.h>
#include "config.h"
#include "ispell.h"
#include "proto.h"


static void chk_suf(ichar_t * word, ichar_t * ucword, int len,
		    int optflags, struct flagent * pfxent, int ignoreflagbits,
		    int allhits);
static void pfx_list_chk(ichar_t * word, ichar_t * ucword,
			 int len, int optflags, int sfxopts, struct flagptr * ind,
			 int ignoreflagbits, int allhits);
static void suf_list_chk(ichar_t * word, ichar_t * ucword, int len,
			 struct flagptr * ind, int optflags, struct flagent * pfxent,
			 int ignoreflagbits, int allhits);

void chk_aff(ichar_t * word, ichar_t * ucword, int len,
	     int ignoreflagbits, int allhits, int pfxopts, int sfxopts);


int good (ichar_t * word, int ignoreflagbits, int allhits,
	  int pfxopts, int sfxopts);
#ifndef NO_CAPITALIZATION_SUPPORT
int cap_ok (ichar_t * word, struct success * hit, int len);
static int entryhasaffixes(struct dent * dent, struct success * hit);
#endif /* NO_CAPITALIZATION_SUPPORT */
void flagpr(ichar_t * word, int preflag, int prestrip,
	    int preadd, int sufflag, int sufadd);


static ichar_t *orig_word;


/* Check possible affixes */
void chk_aff (word, ucword, len, ignoreflagbits, allhits, pfxopts, sfxopts)
     ichar_t *word;		/* Word to be checked */
     ichar_t *ucword;		/* Upper-case-only copy of word */
     int len;		/* The length of word/ucword */
     int ignoreflagbits;	/* Ignore whether affix is legal */
     int allhits;	/* Keep going after first hit */
     int pfxopts;	/* Options to apply to prefixes */
     int sfxopts;	/* Options to apply to suffixes */
{
  register ichar_t *cp;		/* Pointer to char to index on */
  struct flagptr *ind;		/* Flag index table to test */

  pfx_list_chk (word, ucword, len, pfxopts, sfxopts, &pflagindex[0],
		ignoreflagbits, allhits);

  cp = ucword;
  ind = &pflagindex[*cp++];
  while (ind->numents == 0  &&  ind->pu.fp != NULL)
    {
      if (*cp == 0)
	return;
      if (ind->pu.fp[0].numents)
	{

	  pfx_list_chk (word, ucword, len, pfxopts, sfxopts, &ind->pu.fp[0],
			ignoreflagbits, allhits);
	  if (numhits  &&  !allhits  &&  !cflag  &&  !ignoreflagbits)
	    return;
	}
      ind = &ind->pu.fp[*cp++];
    }

  pfx_list_chk (word, ucword, len, pfxopts, sfxopts, ind, ignoreflagbits,
		allhits);

  if (numhits  &&  !allhits  &&  !cflag  &&  !ignoreflagbits)
    return;

  chk_suf (word, ucword, len, sfxopts, (struct flagent *) NULL,
	   ignoreflagbits, allhits);
}

/* Check some prefix flags */
static void pfx_list_chk (word, ucword, len, optflags, sfxopts, ind,
			  ignoreflagbits, allhits)
     ichar_t *		word;		/* Word to be checked */
     ichar_t *		ucword;		/* Upper-case-only word */
     int			len;		/* The length of ucword */
     int			optflags;	/* Options to apply */
     int			sfxopts;	/* Options to apply to suffixes */
     struct flagptr *	ind;		/* Flag index table */
     int			ignoreflagbits;	/* Ignore whether affix is legal */
     int			allhits;	/* Keep going after first hit */
{
  int			cond;		/* Condition number */
  register ichar_t *	cp;		/* Pointer into end of ucword */
  struct dent *	dent;		/* Dictionary entry we found */
  int			entcount;	/* Number of entries to process */
  register struct flagent *
    flent;		/* Current table entry */
  int			preadd;		/* Length added to tword2 as prefix */
  register int	tlen;		/* Length of tword */
  ichar_t		tword[INPUTWORDLEN + 4 * MAXAFFIXLEN + 4]; /* Tmp cpy */
  ichar_t		tword2[sizeof tword]; /* 2nd copy for ins_root_cap */

  for (flent = ind->pu.ent, entcount = ind->numents;
       entcount > 0;
       flent++, entcount--)
    {
      /*
       * If this is a compound-only affix, ignore it unless we're
       * looking for that specific thing.
       */
      if ((flent->flagflags & FF_COMPOUNDONLY) != 0
	  &&  (optflags & FF_COMPOUNDONLY) == 0)
	continue;
      /*
       * In COMPOUND_CONTROLLED mode, the FF_COMPOUNDONLY bit must
       * match exactly.
       */
      if (compoundflag == COMPOUND_CONTROLLED
	  &&  ((flent->flagflags ^ optflags) & FF_COMPOUNDONLY) != 0)
	continue;
      /*
       * See if the prefix matches.
       */
      tlen = len - flent->affl;
      if (tlen > 0
	  &&  (flent->affl == 0
	       ||  icharncmp (flent->affix, ucword, flent->affl) == 0)
	  &&  tlen + flent->stripl >= flent->numconds)
	{
	  /*
	   * The prefix matches.  Remove it, replace it by the "strip"
	   * string (if any), and check the original conditions.
	   */
	  if (flent->stripl)
	    (void) icharcpy (tword, flent->strip);
	  (void) icharcpy (tword + flent->stripl, ucword + flent->affl);
	  cp = tword;
	  for (cond = 0;  cond < flent->numconds;  cond++)
	    {
	      if ((flent->conds[*cp++] & (1 << cond)) == 0)
		break;
	    }
	  if (cond >= flent->numconds)
	    {
	      /*
	       * The conditions match.  See if the word is in the
	       * dictionary.
	       */
	      tlen += flent->stripl;
	      if (cflag)
		flagpr (tword, BITTOCHAR (flent->flagbit), flent->stripl,
			flent->affl, -1, 0);
	      else if (ignoreflagbits)
		{
		  if ((dent = lookup (tword, 1)) != NULL)
		    {
		      cp = tword2;
		      if (flent->affl)
			{
			  (void) icharcpy (cp, flent->affix);
			  cp += flent->affl;
//MS			  *cp++ = '+';
*cp++ = '+';
*cp++ = '\\';
			}
		      preadd = cp - tword2;
		      (void) icharcpy (cp, tword);
		      cp += tlen;
		      if (flent->stripl)
			{
//MS			  *cp++ = '-';
*cp++ = '-';
*cp++ = '\\';
			  (void) icharcpy (cp, flent->strip);
			}
*cp++ = '?';
*cp++ = '\0';

		      (void) ins_root_cap (tword2, word,
					   flent->stripl, preadd,
					   0, (cp - tword2) - tlen - preadd,
					   dent, flent, (struct flagent *) NULL);
		    }
		}
	      else if ((dent = lookup (tword, 1)) != NULL
		       &&  TSTMASKBIT (dent->mask, flent->flagbit))
		{
		  if (numhits < MAX_HITS)
		    {
		      hits[numhits].dictent = dent;
		      hits[numhits].prefix = flent;
		      hits[numhits].suffix = NULL;
		      numhits++;
		    }
		  if (!allhits)
		    {
#ifndef NO_CAPITALIZATION_SUPPORT
		      if (cap_ok (word, &hits[0], len))
			return;
		      numhits = 0;
#else /* NO_CAPITALIZATION_SUPPORT */
		      return;
#endif /* NO_CAPITALIZATION_SUPPORT */
		    }
		}
	      /*
	       * Handle cross-products.
	       */
	      if (flent->flagflags & FF_CROSSPRODUCT)
		chk_suf (word, tword, tlen, sfxopts | FF_CROSSPRODUCT,
			 flent, ignoreflagbits, allhits);
	    }
	}
    }
}

/* Check possible suffixes */
static void chk_suf (word, ucword, len, optflags, pfxent, ignoreflagbits,
		     allhits)
     ichar_t *		word;		/* Word to be checked */
     ichar_t *		ucword;		/* Upper-case-only word */
     int			len;		/* The length of ucword */
     int			optflags;	/* Affix option flags */
     struct flagent *	pfxent;		/* Prefix flag entry if cross-prod */
     int			ignoreflagbits;	/* Ignore whether affix is legal */
     int			allhits;	/* Keep going after first hit */
{
  register ichar_t *	cp;		/* Pointer to char to index on */
  struct flagptr *	ind;		/* Flag index table to test */

  suf_list_chk (word, ucword, len, &sflagindex[0], optflags, pfxent,
		ignoreflagbits, allhits);
  cp = ucword + len - 1;
  ind = &sflagindex[*cp];
  while (ind->numents == 0  &&  ind->pu.fp != NULL)
    {
      if (cp == ucword)
	return;
      if (ind->pu.fp[0].numents)
	{
	  suf_list_chk (word, ucword, len, &ind->pu.fp[0],
			optflags, pfxent, ignoreflagbits, allhits);
	  if (numhits != 0  &&  !allhits  &&  !cflag  &&  !ignoreflagbits)
	    return;
	}
      ind = &ind->pu.fp[*--cp];
    }
  suf_list_chk (word, ucword, len, ind, optflags, pfxent,
		ignoreflagbits, allhits);
}
    
static void suf_list_chk (word, ucword, len, ind, optflags, pfxent,
			  ignoreflagbits, allhits)
     ichar_t *		word;		/* Word to be checked */
     ichar_t *		ucword;		/* Upper-case-only word */
     int			len;		/* The length of ucword */
     struct flagptr *	ind;		/* Flag index table */
     int			optflags;	/* Affix option flags */
     struct flagent *	pfxent;		/* Prefix flag entry if crossonly */
     int			ignoreflagbits;	/* Ignore whether affix is legal */
     int			allhits;	/* Keep going after first hit */
{
  register ichar_t *	cp;		/* Pointer into end of ucword */
  int			cond;		/* Condition number */
  struct dent *	dent;		/* Dictionary entry we found */
  int			entcount;	/* Number of entries to process */
  register struct flagent *
    flent;		/* Current table entry */
  int			preadd;		/* Length added to tword2 as prefix */
  register int	tlen;		/* Length of tword */
  ichar_t		tword[INPUTWORDLEN + 4 * MAXAFFIXLEN + 4]; /* Tmp cpy */
  ichar_t		tword2[sizeof tword]; /* 2nd copy for ins_root_cap */

  (void) icharcpy (tword, ucword);
  for (flent = ind->pu.ent, entcount = ind->numents;
       entcount > 0;
       flent++, entcount--)
    {
      if ((optflags & FF_CROSSPRODUCT) != 0
	  &&  (flent->flagflags & FF_CROSSPRODUCT) == 0)
	continue;
      /*
       * If this is a compound-only affix, ignore it unless we're
       * looking for that specific thing.
       */
      if ((flent->flagflags & FF_COMPOUNDONLY) != 0
	  &&  (optflags & FF_COMPOUNDONLY) == 0)
	continue;
      /*
       * In COMPOUND_CONTROLLED mode, the FF_COMPOUNDONLY bit must
       * match exactly.
       */
      if (compoundflag == COMPOUND_CONTROLLED
	  &&  ((flent->flagflags ^ optflags) & FF_COMPOUNDONLY) != 0)
	continue;
      /*
       * See if the suffix matches.
       */
      tlen = len - flent->affl;
      if (tlen > 0
	  &&  (flent->affl == 0
	       ||  icharcmp (flent->affix, ucword + tlen) == 0)
	  &&  tlen + flent->stripl >= flent->numconds)
	{
	  /*
	   * The suffix matches.  Remove it, replace it by the "strip"
	   * string (if any), and check the original conditions.
	   */
	  (void) icharcpy (tword, ucword);
	  cp = tword + tlen;
	  if (flent->stripl)
	    {
	      (void) icharcpy (cp, flent->strip);
	      tlen += flent->stripl;
	      cp = tword + tlen;
	    }
	  else
	    *cp = '\0';
	  for (cond = flent->numconds;  --cond >= 0;  )
	    {
	      if ((flent->conds[*--cp] & (1 << cond)) == 0)
		break;
	    }
	  if (cond < 0)
	    {
	      /*
	       * The conditions match.  See if the word is in the
	       * dictionary.
	       */
	      if (cflag)
		{
		  if (optflags & FF_CROSSPRODUCT)
		    flagpr (tword, BITTOCHAR (pfxent->flagbit),
			    pfxent->stripl, pfxent->affl,
			    BITTOCHAR (flent->flagbit), flent->affl);
		  else
		    flagpr (tword, -1, 0, 0,
			    BITTOCHAR (flent->flagbit), flent->affl);
		}
	      else if (ignoreflagbits)
		{
		  if ((dent = lookup (tword, 1)) != NULL)
		    {
		      cp = tword2;
		      if ((optflags & FF_CROSSPRODUCT)
			  &&  pfxent->affl != 0)
			{
			  (void) icharcpy (cp, pfxent->affix);
			  cp += pfxent->affl;
//MS			  *cp++ = '+';
*cp++ = '+';
*cp++ = '\\';
			}
		      preadd = cp - tword2;
		      (void) icharcpy (cp, tword);
		      cp += tlen;
		      if ((optflags & FF_CROSSPRODUCT)
			  &&  pfxent->stripl != 0)
			{
//MS			  *cp++ = '-';
*cp++ = '-';
*cp++ = '\\';
			  (void) icharcpy (cp, pfxent->strip);
			  cp += pfxent->stripl;
			}
		      if (flent->stripl)
			{
//MS			  *cp++ = '-';
*cp++ = '-';
*cp++ = '\\';
			  (void) icharcpy (cp, flent->strip);
			  cp += flent->stripl;
			}
		      if (flent->affl)
			{
//MS			  *cp++ = '+';
*cp++ = '+';
*cp++ = '\\';
			  (void) icharcpy (cp, flent->affix);
			  cp += flent->affl;
			}
*cp++ = '?';
*cp++ = '\0';
		      (void) ins_root_cap (tword2, word,
					   (optflags & FF_CROSSPRODUCT) ? pfxent->stripl : 0,
					   preadd,
					   flent->stripl, (cp - tword2) - tlen - preadd,
					   dent, pfxent, flent);
		    }
		}
	      else if ((dent = lookup (tword, 1)) != NULL
		       &&  TSTMASKBIT (dent->mask, flent->flagbit)
		       &&  ((optflags & FF_CROSSPRODUCT) == 0
			    || TSTMASKBIT (dent->mask, pfxent->flagbit)))
		{
		  if (numhits < MAX_HITS)
		    {
		      hits[numhits].dictent = dent;
		      hits[numhits].prefix = pfxent;
		      hits[numhits].suffix = flent;
		      numhits++;
		    }
		  if (!allhits)
		    {
#ifndef NO_CAPITALIZATION_SUPPORT
		      if (cap_ok (word, &hits[0], len))
			return;
		      numhits = 0;
#else /* NO_CAPITALIZATION_SUPPORT */
		      return;
#endif /* NO_CAPITALIZATION_SUPPORT */
		    }
		}
	    }
	}
    }
}


#ifndef NO_CAPITALIZATION_SUPPORT
int good (w, ignoreflagbits, allhits, pfxopts, sfxopts)
     ichar_t *w;		/* Word to look up */
     int ignoreflagbits;	/* NZ to ignore affix flags in dict */
     int allhits;	/* NZ to ignore case, get every hit */
     int pfxopts;	/* Options to apply to prefixes */
     int sfxopts;	/* Options to apply to suffixes */
#else
     /* ARGSUSED */
     int good (w, ignoreflagbits, dummy, pfxopts, sfxopts)
     ichar_t *w;		/* Word to look up */
     int ignoreflagbits;	/* NZ to ignore affix flags in dict */
     int dummy;
#define allhits	0	/* Never actually need more than one hit */
     int pfxopts;	/* Options to apply to prefixes */
     int sfxopts;	/* Options to apply to suffixes */
#endif
{
  ichar_t		nword[INPUTWORDLEN + MAXAFFIXLEN];
  register ichar_t *	p;
  register ichar_t *	q;
  register int		n;
  register struct dent * dp;

  /*
  ** Make an uppercase copy of the word we are checking.
  */
  for (p = w, q = nword;  *p;  )
    *q++ = mytoupper (*p++);
  *q = 0;
  n = q - nword;

  numhits = 0;

  if ((dp = lookup (nword, 1)) != NULL)
    {
      hits[0].dictent = dp;
      hits[0].prefix = NULL;
      hits[0].suffix = NULL;
#ifndef NO_CAPITALIZATION_SUPPORT
      if (allhits  ||  cap_ok (w, &hits[0], n))
	numhits = 1;
#else
      numhits = 1;
#endif

      /*
       * If we're looking for compounds, and this root doesn't
       * participate in compound formation, undo the hit.
       */
      if (compoundflag == COMPOUND_CONTROLLED
	  &&  ((pfxopts | sfxopts) & FF_COMPOUNDONLY) != 0
	  &&  hashheader.compoundbit >= 0
	  &&  TSTMASKBIT (dp->mask, hashheader.compoundbit) == 0)
	numhits = 0;
    }

  if (numhits  &&  !allhits)
    return 1;

  /* try stripping off affixes */

  chk_aff (w, nword, n, ignoreflagbits, allhits, pfxopts, sfxopts);

  return numhits;
}

#ifndef NO_CAPITALIZATION_SUPPORT
int cap_ok (word, hit, len)
     register ichar_t *		word;
     register struct success *	hit;
     int				len;
{
  register ichar_t *		dword;
  register ichar_t *		w;
  register struct dent *	dent;
  ichar_t			dentword[INPUTWORDLEN + MAXAFFIXLEN];
  int				preadd;
  int				prestrip;
  int				sufadd;
  ichar_t *			limit;
  long			thiscap;
  long			dentcap;

  thiscap = whatcap (word);
  /*
  ** All caps is always legal, regardless of affixes.
  */
  preadd = prestrip = sufadd = 0;
  if (thiscap == ALLCAPS)
    return 1;
  else if (thiscap == FOLLOWCASE)
    {
      /* Set up some constants for the while(1) loop below */
      if (hit->prefix)
	{
	  preadd = hit->prefix->affl;
	  prestrip = hit->prefix->stripl;
	}
      else
	preadd = prestrip = 0;
      sufadd = hit->suffix ? hit->suffix->affl : 0;
    }
  /*
  ** Search the variants for one that matches what we have.  Note
  ** that thiscap can't be ALLCAPS, since we already returned
  ** for that case.
  */
  dent = hit->dictent;
  for (  ;  ;  )
    {
      dentcap = captype (dent->flagfield);
      if (dentcap != thiscap)
	{
	  if (dentcap == ANYCASE  &&  thiscap == CAPITALIZED
	      &&  entryhasaffixes (dent, hit))
	    return 1;
	}
      else				/* captypes match */
	{
	  if (thiscap != FOLLOWCASE)
	    {
	      if (entryhasaffixes (dent, hit))
		return 1;
	    }
	  else
	    {
	      /*
	      ** Make sure followcase matches exactly.
	      ** Life is made more difficult by the
	      ** possibility of affixes.  Start with
	      ** the prefix.
	      */
	      (void) strtoichar (dentword, dent->word, INPUTWORDLEN, 1);
	      dword = dentword;
	      limit = word + preadd;
	      if (myupper (dword[prestrip]))
		{
		  for (w = word;  w < limit;  w++)
		    {
		      if (mylower (*w))
			goto doublecontinue;
		    }
		}
	      else
		{
		  for (w = word;  w < limit;  w++)
		    {
		      if (myupper (*w))
			goto doublecontinue;
		    }
		}
	      dword += prestrip;
	      /* Do root part of word */
	      limit = dword + len - preadd - sufadd;
	      while (dword < limit)
		{
		  if (*dword++ != *w++)
		    goto doublecontinue;
		}
	      /* Do suffix */
	      dword = limit - 1;
	      if (myupper (*dword))
		{
		  for (  ;  *w;  w++)
		    {
		      if (mylower (*w))
			goto doublecontinue;
		    }
		}
	      else
		{
		  for (  ;  *w;  w++)
		    {
		      if (myupper (*w))
			goto doublecontinue;
		    }
		}
	      /*
	      ** All failure paths go to "doublecontinue,"
	      ** so if we get here it must match.
	      */
	      if (entryhasaffixes (dent, hit))
		return 1;
	    doublecontinue:	;
	    }
	}
      if ((dent->flagfield & MOREVARIANTS) == 0)
	break;
      dent = dent->next;
    }

  /* No matches found */
  return 0;
}

/*
** See if this particular capitalization (dent) is legal with these
** particular affixes.
*/
static int entryhasaffixes (dent, hit)
     register struct dent *	dent;
     register struct success *	hit;
{

  if (hit->prefix  &&  !TSTMASKBIT (dent->mask, hit->prefix->flagbit))
    return 0;
  if (hit->suffix  &&  !TSTMASKBIT (dent->mask, hit->suffix->flagbit))
    return 0;
  return 1;			/* Yes, these affixes are legal */
}
#endif

/*
 * Print a word and its flag, making sure the case of the output matches
 * the case of the original found in "orig_word".
 */
void flagpr (word, preflag, prestrip, preadd, sufflag, sufadd)
     register ichar_t *	word;		/* (Modified) word to print */
     int			preflag;	/* Prefix flag (if any) */
     int			prestrip;	/* Lth of pfx stripped off orig_word */
     int			preadd;		/* Length of prefix added to w */
     int			sufflag;	/* Suffix flag (if any) */
     int			sufadd;		/* Length of suffix added to w */
{
  register ichar_t *	origp;		/* Pointer into orig_word */
  int			orig_len;	/* Length of orig_word */

  orig_len = icharlen (orig_word);
  /*
   * We refuse to print if the cases outside the modification
   * points don't match those just inside.  This prevents things
   * like "OEM's" from being turned into "OEM/S" which expands
   * only to "OEM'S".
   */
  if (preflag > 0)
    {
      origp = orig_word + preadd;
      if (myupper (*origp))
	{
	  for (origp = orig_word;  origp < orig_word + preadd;  origp++)
	    {
	      if (mylower (*origp))
		return;
	    }
	}
      else
	{
	  for (origp = orig_word;  origp < orig_word + preadd;  origp++)
	    {
	      if (myupper (*origp))
		return;
	    }
	}
    }
  if (sufflag > 0)
    {
      origp = orig_word + orig_len - sufadd;
      if (myupper (origp[-1]))
	{
	  for (  ;  *origp != 0;  origp++)
	    {
	      if (mylower (*origp))
		return;
	    }
	}
      else
	{
	  origp = orig_word + orig_len - sufadd;
	  for (  ;  *origp != 0;  origp++)
	    {
	      if (myupper (*origp))
		return;
	    }
	}
    }
  /*
   * The cases are ok.  Put out the word, being careful that the
   * prefix/suffix cases match those in the original, and that the
   * unchanged characters from the original actually match it.
   */
  (void) putchar (' ');
  origp = orig_word + preadd;
  if (myupper (*origp))
    {
      while (--prestrip >= 0)
	(void) fputs (printichar ((int) *word++), stdout);
    }
  else
    {
      while (--prestrip >= 0)
	(void) fputs (printichar ((int) mytolower (*word++)), stdout);
    }
  for (prestrip = orig_len - preadd - sufadd;  --prestrip >= 0;  word++)
    (void) fputs (printichar ((int) *origp++), stdout);
  if (origp > orig_word)
    origp--;
  if (myupper (*origp))
    (void) fputs (ichartosstr (word, 0), stdout);
  else
    {
      while (*word)
	{
	  (void) fputs (printichar ((int) mytolower (*word++)), stdout);
	}
    }
  /*
   * Now put out the flags
   */
  (void) putchar (hashheader.flagmarker);
  if (preflag > 0)
    (void) putchar (preflag);
  if (sufflag > 0)
    (void) putchar (sufflag);
}
